"""Utility functions for Dapla Pseudo."""

import io
import os
import typing as t
import zipfile
from datetime import date
from pathlib import Path

import fsspec
import orjson
import polars as pl
from dapla import FileClient
from gcsfs.core import GCSFile
from google.auth.exceptions import DefaultCredentialsError

from dapla_pseudo.constants import PseudoOperation
from dapla_pseudo.exceptions import FileInvalidError
from dapla_pseudo.exceptions import MimetypeNotSupportedError
from dapla_pseudo.exceptions import NoFileExtensionError
from dapla_pseudo.types import BinaryFileDecl
from dapla_pseudo.types import FileLikeDatasetDecl
from dapla_pseudo.v1.models.api import DepseudoFieldRequest
from dapla_pseudo.v1.models.api import DepseudoFileRequest
from dapla_pseudo.v1.models.api import PseudoFieldRequest
from dapla_pseudo.v1.models.api import PseudoFileRequest
from dapla_pseudo.v1.models.api import RepseudoFieldRequest
from dapla_pseudo.v1.models.api import RepseudoFileRequest
from dapla_pseudo.v1.models.core import KeyWrapper
from dapla_pseudo.v1.models.core import Mimetypes
from dapla_pseudo.v1.models.core import PseudoConfig
from dapla_pseudo.v1.models.core import PseudoKeyset
from dapla_pseudo.v1.models.core import PseudoRule
from dapla_pseudo.v1.mutable_dataframe import MutableDataFrame
from dapla_pseudo.v1.supported_file_format import FORMAT_TO_MIMETYPE_FUNCTION
from dapla_pseudo.v1.supported_file_format import SupportedOutputFileFormat


def find_multipart_obj(obj_name: str, multipart_files_tuple: set[t.Any]) -> t.Any:
    """Find "multipart object" by name.

    The requests lib specifies multipart file arguments as file-tuples, such as
    ('filename', fileobj, 'content_type')
    This method searches a tuple of such file-tuples ((file-tuple1),...,(file-tupleN))
    It returns the fileobj for the first matching file-tuple with a specified filename.

    Example:
    Given the multipart_files_tuple:
    multipart_tuple = (('filename1', fileobj1, 'application/json'), ('filename2', fileobj2, 'application/json'))

    then
    find_multipart_obj("filename2", multipart_tuple) -> fileobj
    """
    try:
        matching_item = next(
            item[1] for item in multipart_files_tuple if item[0] == obj_name
        )
        return matching_item[1]
    except StopIteration:
        return None


def convert_to_date(sid_snapshot_date: date | str | None = None) -> date | None:
    """Converts the SID version date to the 'date' type, if it is a string.

    If None, simply passes the None through the function.
    """
    if isinstance(sid_snapshot_date, str):
        try:
            return date.fromisoformat(sid_snapshot_date)
        except ValueError as exc:
            raise ValueError(
                "Version timestamp must be a valid ISO date string (YYYY-MM-DD)"
            ) from exc
    return sid_snapshot_date


def get_file_format_from_file_name(file_path: str | Path) -> SupportedOutputFileFormat:
    """Extracts the file format from a file path."""
    if isinstance(file_path, str):
        file_path = Path(file_path)

    file_extension = file_path.suffix
    if not file_extension:
        raise NoFileExtensionError(f"File path '{file_path}' has no file extension.")
    file_format = SupportedOutputFileFormat(file_extension.replace(".", ""))
    return file_format


def build_pseudo_field_request(
    pseudo_operation: PseudoOperation,
    mutable_df: MutableDataFrame,
    rules: list[PseudoRule],  # "source rules" if repseudo
    custom_keyset: PseudoKeyset | str | None = None,
    target_custom_keyset: PseudoKeyset | str | None = None,  # used in repseudo
    target_rules: list[PseudoRule] | None = None,  # used in repseudo)
) -> list[PseudoFieldRequest | DepseudoFieldRequest | RepseudoFieldRequest]:
    """Builds a FieldRequest object."""
    mutable_df.match_rules(rules, target_rules)
    matched_fields = mutable_df.get_matched_fields()
    match pseudo_operation:
        case PseudoOperation.PSEUDONYMIZE:
            return [
                PseudoFieldRequest(
                    pseudo_func=field.func,
                    name=field.path,
                    pattern=field.pattern,
                    values=field.col["values"],
                    keyset=KeyWrapper(custom_keyset).keyset,
                )
                for field in matched_fields
            ]
        case PseudoOperation.DEPSEUDONYMIZE:
            return [
                DepseudoFieldRequest(
                    pseudo_func=field.func,
                    name=field.path,
                    pattern=field.pattern,
                    values=field.col["values"],
                    keyset=KeyWrapper(custom_keyset).keyset,
                )
                for field in matched_fields
            ]
        case PseudoOperation.REPSEUDONYMIZE:
            if target_rules is not None:
                return [
                    RepseudoFieldRequest(
                        source_pseudo_func=field.func,
                        target_pseudo_func=field.target_func,
                        name=field.path,
                        pattern=field.pattern,
                        values=field.col["values"],
                        source_keyset=KeyWrapper(custom_keyset).keyset,
                        target_keyset=KeyWrapper(target_custom_keyset).keyset,
                    )
                    for field in matched_fields
                ]
            else:
                raise ValueError("Found no target rules")


def build_pseudo_file_request(
    pseudo_operation: PseudoOperation,
    rules: list[PseudoRule],  # "source rules" if repseudo
    custom_keyset: PseudoKeyset | str | None = None,
    target_custom_keyset: PseudoKeyset | str | None = None,  # used in repseudo
    target_rules: list[PseudoRule] | None = None,  # used in repseudo)
) -> PseudoFileRequest | DepseudoFileRequest | RepseudoFileRequest:
    """Builds a file request object."""
    match pseudo_operation:
        case PseudoOperation.PSEUDONYMIZE:
            return PseudoFileRequest(
                pseudo_config=PseudoConfig(
                    rules=rules,
                    keysets=KeyWrapper(custom_keyset).keyset_list(),
                ),
                target_content_type=Mimetypes.JSON,
                target_uri=None,
                compression=None,
            )
        case PseudoOperation.DEPSEUDONYMIZE:
            return DepseudoFileRequest(
                pseudo_config=PseudoConfig(
                    rules=rules,
                    keysets=KeyWrapper(custom_keyset).keyset_list(),
                ),
                target_content_type=Mimetypes.JSON,
                target_uri=None,
                compression=None,
            )
        case PseudoOperation.REPSEUDONYMIZE:
            if target_rules is not None:
                return RepseudoFileRequest(
                    source_pseudo_config=PseudoConfig(
                        rules=rules,
                        keysets=KeyWrapper(custom_keyset).keyset_list(),
                    ),
                    target_pseudo_config=PseudoConfig(
                        rules=target_rules,
                        keysets=KeyWrapper(target_custom_keyset).keyset_list(),
                    ),
                    target_content_type=Mimetypes.JSON,
                    target_uri=None,
                    compression=None,
                )
            else:
                raise ValueError("Found no target rules")


def get_file_data_from_dataset(
    dataset: FileLikeDatasetDecl | pl.DataFrame,
) -> tuple[BinaryFileDecl, Mimetypes]:
    """Converts the given dataset to a file handle and content type.

    Args:
        dataset (FileLikeDatasetDecl): The provided dataset

    Raises:
        FileNotFoundError: If the file cannot be found
        DefaultCredentialsError: If the provided dataset is a GCS path and no Google credentials are found
        ValueError: If the provided dataset is not of a supported type
        FileInvalidError: If the file is empty

    Returns:
        tuple[BinaryFileDecl, Mimetypes]: A tuple of (file handle, content type)
    """
    file_handle: BinaryFileDecl | None = None
    match dataset:
        case str() | Path():
            # File path
            if str(dataset).startswith("gs://"):
                try:
                    file_handle = FileClient().gcs_open(str(dataset), mode="rb")
                except OSError as err:
                    raise FileNotFoundError(
                        f"No GCS file found or authentication not sufficient for: {dataset}"
                    ) from err
                except DefaultCredentialsError as err:
                    raise DefaultCredentialsError(
                        "No Google Authentication found in environment"
                    ) from err
            else:
                file_handle = open(dataset, "rb")

            file_handle.seek(0)

        # Convert Polars dataframe to a zipped archive with json data
        case pl.DataFrame() as df:
            file_handle = io.BytesIO()
            with zipfile.ZipFile(
                file_handle, "a", compression=zipfile.ZIP_DEFLATED, compresslevel=9
            ) as zip_file:
                zip_file.writestr("data.json", orjson.dumps(df.to_dicts()))
                zip_file.filename = "data.zip"
            file_handle.seek(0)
            return file_handle, Mimetypes.ZIP

        case io.BufferedReader():
            # File handle
            dataset.seek(0)
            file_handle = dataset
        case fsspec.spec.AbstractBufferedFile():
            # This is a file handle to a remote storage system such as GCS.
            # It provides random access for the underlying file-like data (without downloading the whole thing).
            dataset.seek(0)
            file_handle = io.BufferedReader(dataset)
        case _:
            raise ValueError(
                f"Unsupported data type: {type(dataset)}. Supported types are {FileLikeDatasetDecl}"
            )

    if isinstance(file_handle, GCSFile):
        file_size = file_handle.size
    else:
        file_size = os.fstat(file_handle.fileno()).st_size

    if file_size == 0:
        raise FileInvalidError("File is empty.")

    content_type = get_content_type_from_file(file_handle)
    return file_handle, content_type


def get_content_type_from_file(file_handle: BinaryFileDecl) -> Mimetypes:
    """Given a file handle, extracts the content type from the file.

    Args:
        file_handle (BinaryFileDecl): The file handle to extract the content type from.

    Raises:
        MimetypeNotSupportedError: If the Mimetype is not supported

    Returns:
        Mimetypes: The Mimetype of the file.
    """
    if isinstance(file_handle, GCSFile):
        file_name = file_handle.full_name
    else:
        file_name = file_handle.name
    file_format = get_file_format_from_file_name(file_name)
    try:  # Test whether the file extension is supported
        content_type = Mimetypes(FORMAT_TO_MIMETYPE_FUNCTION[file_format])
    except KeyError:
        raise MimetypeNotSupportedError(
            f"The provided input format '{file_format}' is not supported from file."
        ) from None

    return content_type



"""Common exceptions for the Dapla Pseudo package."""


class NoFileExtensionError(Exception):
    """Exception raised when a file has no file extension."""

    def __init__(self, message: str) -> None:
        """Initialize the NoFileExtensionError."""
        super().__init__(message)


class ExtensionNotValidError(Exception):
    """Exception raised when a file extension is invalid."""

    def __init__(self, message: str) -> None:
        """Initialize the ExtensionNotValidError."""
        super().__init__(message)


class MimetypeNotSupportedError(Exception):
    """Exception raised when a Mimetype is invalid."""

    def __init__(self, message: str) -> None:
        """Initialize the ExtensionNotValidError."""
        super().__init__(message)


class FileInvalidError(Exception):
    """Exception raised when a file is in an invalid state."""

    def __init__(self, message: str) -> None:
        """Initialize the NoFileExtensionError."""
        super().__init__(message)



"""Baseclasses for the Pseudo Builders.

The methods are kept private in order to not expose them to users of the client
when using autocomplete-features. The method names should also be more technical
and descriptive than the user-friendly methods that are exposed.
"""

import json
import os
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed
from datetime import date
from typing import cast

import polars as pl
import requests

from dapla_pseudo.constants import Env
from dapla_pseudo.constants import MapFailureStrategy
from dapla_pseudo.constants import PredefinedKeys
from dapla_pseudo.constants import PseudoFunctionTypes
from dapla_pseudo.constants import PseudoOperation
from dapla_pseudo.types import FileSpecDecl
from dapla_pseudo.utils import build_pseudo_field_request
from dapla_pseudo.utils import build_pseudo_file_request
from dapla_pseudo.utils import convert_to_date
from dapla_pseudo.v1.client import PseudoClient
from dapla_pseudo.v1.client import _extract_name
from dapla_pseudo.v1.models.api import DepseudoFieldRequest
from dapla_pseudo.v1.models.api import DepseudoFileRequest
from dapla_pseudo.v1.models.api import PseudoFieldRequest
from dapla_pseudo.v1.models.api import PseudoFieldResponse
from dapla_pseudo.v1.models.api import PseudoFileRequest
from dapla_pseudo.v1.models.api import PseudoFileResponse
from dapla_pseudo.v1.models.api import RawPseudoMetadata
from dapla_pseudo.v1.models.api import RepseudoFieldRequest
from dapla_pseudo.v1.models.api import RepseudoFileRequest
from dapla_pseudo.v1.models.core import DaeadKeywordArgs
from dapla_pseudo.v1.models.core import FF31KeywordArgs
from dapla_pseudo.v1.models.core import File
from dapla_pseudo.v1.models.core import MapSidKeywordArgs
from dapla_pseudo.v1.models.core import Mimetypes
from dapla_pseudo.v1.models.core import PseudoFunction
from dapla_pseudo.v1.models.core import PseudoKeyset
from dapla_pseudo.v1.models.core import PseudoRule
from dapla_pseudo.v1.mutable_dataframe import MutableDataFrame
from dapla_pseudo.v1.result import Result


class _BasePseudonymizer:
    """Base class for the _Pseudonymizer/_Depseudonymizer/_Repseudonymizer builders."""

    def __init__(
        self, pseudo_operation: PseudoOperation, dataset: File | pl.DataFrame
    ) -> None:
        """The constructor of the base class."""
        self._pseudo_operation = pseudo_operation
        self._pseudo_client: PseudoClient = PseudoClient(
            pseudo_service_url=os.getenv(Env.PSEUDO_SERVICE_URL),
            auth_token=os.getenv(Env.PSEUDO_SERVICE_AUTH_TOKEN),
        )
        self._dataset: File | MutableDataFrame
        match dataset:  # Differentiate between file and DataFrame
            case pl.DataFrame():
                self._dataset = MutableDataFrame(dataset)
            case File():
                self._dataset = dataset
            case _ as invalid_dataset:
                raise ValueError(
                    f"Unsupported data type: {type(invalid_dataset)}. Should only be DataFrame or file-like type."
                )

    def _execute_pseudo_operation(
        self,
        rules: list[PseudoRule],  # "source rules" if repseudo
        timeout: int,
        custom_keyset: PseudoKeyset | str | None = None,
        target_custom_keyset: PseudoKeyset | str | None = None,  # used in repseudo
        target_rules: list[PseudoRule] | None = None,  # used in repseudo
    ) -> Result:
        if self._dataset is None:
            raise ValueError("No dataset has been provided.")

        if rules == []:
            raise ValueError(
                "No fields have been provided. Use the 'on_fields' method."
            )

        pseudo_response: PseudoFileResponse | PseudoFieldResponse
        match self._dataset:  # Differentiate between file and DataFrame
            case MutableDataFrame():
                pseudo_requests = build_pseudo_field_request(
                    self._pseudo_operation,
                    self._dataset,
                    rules,
                    custom_keyset,
                    target_custom_keyset,
                    target_rules,
                )
                pseudo_response = self._pseudonymize_field(pseudo_requests, timeout)
                pseudo_response.data = self._dataset.to_polars()
            case File():
                pseudo_request = build_pseudo_file_request(
                    self._pseudo_operation,
                    rules,
                    custom_keyset,
                    target_custom_keyset,
                    target_rules,
                )
                pseudo_response = self._pseudonymize_file(pseudo_request, timeout)
            case _ as invalid_dataset:
                raise ValueError(
                    f"Unsupported data type: {type(invalid_dataset)}. Should only be DataFrame or file-like type."
                )
        return Result(pseudo_response=pseudo_response)

    def _pseudonymize_field(
        self,
        pseudo_requests: list[
            PseudoFieldRequest | DepseudoFieldRequest | RepseudoFieldRequest
        ],
        timeout: int,
    ) -> PseudoFieldResponse:
        """Pseudonymizes the specified fields in the DataFrame using the provided pseudonymization function.

        The pseudonymization is performed in parallel. After the parallel processing is finished,
        the pseudonymized fields replace the original fields in the DataFrame stored in `self._dataframe`.
        """

        def pseudonymize_field_runner(
            request: PseudoFieldRequest | DepseudoFieldRequest | RepseudoFieldRequest,
        ) -> tuple[str, list[str], RawPseudoMetadata]:
            """Function that performs the pseudonymization on a Polars Series."""
            response: requests.Response = self._pseudo_client._post_to_field_endpoint(
                f"{self._pseudo_operation.value}/field",
                request,
                timeout,
                stream=True,
            )
            payload = json.loads(response.content.decode("utf-8"))
            data = payload["data"]
            metadata = RawPseudoMetadata(
                field_name=request.name,
                logs=payload["logs"],
                metrics=payload["metrics"],
                datadoc=payload["datadoc_metadata"]["pseudo_variables"],
            )

            return request.name, data, metadata

        # type narrowing isn't carried over from caller function
        assert isinstance(self._dataset, MutableDataFrame)
        # Execute the pseudonymization API calls in parallel
        with ThreadPoolExecutor() as executor:
            raw_metadata_fields: list[RawPseudoMetadata] = []
            futures = [
                executor.submit(pseudonymize_field_runner, request)
                for request in pseudo_requests
            ]

            for future in as_completed(futures):
                field_name, data, raw_metadata = future.result()
                self._dataset.update(field_name, data)
                raw_metadata_fields.append(raw_metadata)

            return PseudoFieldResponse(
                data=self._dataset.to_polars(), raw_metadata=raw_metadata_fields
            )

    def _pseudonymize_file(
        self,
        pseudo_request: PseudoFileRequest | DepseudoFileRequest | RepseudoFileRequest,
        timeout: int,
    ) -> PseudoFileResponse:
        self._dataset = cast(File, self._dataset)
        file_handle = self._dataset.file_handle
        content_type = self._dataset.content_type
        request_spec: FileSpecDecl = (
            None,
            pseudo_request.to_json(),
            str(Mimetypes.JSON),
        )

        file_name = _extract_name(
            file_handle=file_handle, input_content_type=content_type
        )
        data_spec = (
            file_name,
            file_handle,
            str(pseudo_request.target_content_type),
        )

        response = self._pseudo_client._post_to_file_endpoint(
            path=f"{self._pseudo_operation.value}/file",
            request_spec=request_spec,
            data_spec=data_spec,
            stream=True,
            timeout=timeout,
        )
        file_handle.close()

        payload = json.loads(response.content.decode("utf-8"))
        pseudo_data = payload["data"]
        metadata = RawPseudoMetadata(
            logs=payload["logs"],
            metrics=payload["metrics"],
            datadoc=payload["datadoc_metadata"]["pseudo_variables"],
        )

        return PseudoFileResponse(
            data=pseudo_data,
            raw_metadata=metadata,
            content_type=Mimetypes.JSON,
            streamed=True,
            file_name=file_name,
        )


class _BaseRuleConstructor:
    """Base class for the _PseudoFuncSelector/_DepseudoFuncSelector/_RepseudoFuncSelector builders."""

    def __init__(
        self,
        fields: list[str],
        dataset_type: type[pl.DataFrame] | type[File],
    ) -> None:
        self._fields = fields
        self._dataset_type = dataset_type

    def _map_to_stable_id_and_pseudonymize(
        self,
        sid_snapshot_date: str | date | None = None,
        custom_key: PredefinedKeys | str | None = None,
        on_map_failure: MapFailureStrategy | str | None = None,
    ) -> list[PseudoRule]:
        failure_strategy = (
            None if on_map_failure is None else MapFailureStrategy(on_map_failure)
        )
        kwargs = (
            MapSidKeywordArgs(
                key_id=custom_key,
                snapshot_date=convert_to_date(sid_snapshot_date),
                failure_strategy=failure_strategy,
            )
            if custom_key
            else MapSidKeywordArgs(
                snapshot_date=convert_to_date(sid_snapshot_date),
                failure_strategy=failure_strategy,
            )
        )
        pseudo_func = PseudoFunction(
            function_type=PseudoFunctionTypes.MAP_SID, kwargs=kwargs
        )
        return self._rule_constructor(pseudo_func)

    def _with_daead_encryption(
        self, custom_key: PredefinedKeys | str | None = None
    ) -> list[PseudoRule]:
        kwargs = (
            DaeadKeywordArgs(key_id=custom_key) if custom_key else DaeadKeywordArgs()
        )
        pseudo_func = PseudoFunction(
            function_type=PseudoFunctionTypes.DAEAD, kwargs=kwargs
        )
        return self._rule_constructor(pseudo_func)

    def _with_ff31_encryption(
        self, custom_key: PredefinedKeys | str | None = None
    ) -> list[PseudoRule]:
        kwargs = FF31KeywordArgs(key_id=custom_key) if custom_key else FF31KeywordArgs()
        pseudo_func = PseudoFunction(
            function_type=PseudoFunctionTypes.FF31, kwargs=kwargs
        )
        return self._rule_constructor(pseudo_func)

    def _with_custom_function(self, function: PseudoFunction) -> list[PseudoRule]:
        return self._rule_constructor(function)

    def _rule_constructor(self, func: PseudoFunction) -> list[PseudoRule]:
        # If we use the pseudonymize_file endpoint, we need a glob catch-all prefix.
        rule_prefix = "**/" if self._dataset_type == File else ""
        rules = [
            PseudoRule(name=None, func=func, pattern=f"{rule_prefix}{field}")
            for field in self._fields
        ]
        return rules



import re
import typing as t
from collections import Counter
from functools import cache
from io import BytesIO

import orjson
import polars as pl
from wcmatch import glob

from dapla_pseudo.v1.models.core import PseudoFunction
from dapla_pseudo.v1.models.core import PseudoRule

ARRAY_INDEX_MATCHER = re.compile(r"\[\d*]")


def _ensure_normalized(pattern: str) -> str:
    """Normalize the pattern.

    Ensure that the pattern always starts with a '/' or '*' to be compatible with the
    pattern matching that is used in pseudo-service.
    """
    return (
        pattern
        if (pattern.startswith("/") or pattern.startswith("*"))
        else "/" + pattern
    )


class FieldMatch:
    """Represents a reference to a matching column in the dataframe."""

    def __init__(
        self,
        path: str,
        pattern: str,
        col: dict[str, t.Any],
        func: PseudoFunction,  # "source func" if repseudo
        target_func: PseudoFunction | None,  # "target_func" if repseudo
    ) -> None:
        """Initialize the class."""
        self.path = path
        self.pattern = _ensure_normalized(pattern)
        self.col = col
        self.func = func
        self.target_func = target_func

    def update_col(self, key: str, data: list[str]) -> None:
        """Update the values in the matched column."""
        self.col.update({key: data})


class MutableDataFrame:
    """A dataframe that can change values in-place."""

    def __init__(self, dataframe: pl.DataFrame) -> None:
        """Initialize the class."""
        self.dataframe = dataframe
        self.dataframe_dict: t.Any = None
        self.matched_fields: list[FieldMatch] = []
        self.matched_fields_metrics: dict[str, int] | None = None

    def match_rules(
        self, rules: list[PseudoRule], target_rules: list[PseudoRule] | None
    ) -> None:
        """Create references to all the columns that matches the given pseudo rules."""
        counter: Counter[str] = Counter()
        self.dataframe_dict = orjson.loads(self.dataframe.write_json())
        self.matched_fields = list(
            _traverse_dataframe_dict(
                self.dataframe_dict["columns"],
                _combine_rules(rules, target_rules),
                counter,
            )
        )
        # The Counter contains unique field names. A count > 1 means that the traverse
        # was not able to group all values with a given path. This will be the case for
        # list of dicts.
        self.matched_fields_metrics = dict(counter)

    def get_matched_fields(self) -> list[FieldMatch]:
        """Get a reference to all the columns that matched pseudo rules."""
        return self.matched_fields

    def update(self, field: str, data: list[str]) -> None:
        """Update a column with the given data."""
        if any((field_match := f) for f in self.matched_fields if field == f.path):
            field_match.update_col("values", data)

    def to_polars(self) -> pl.DataFrame:
        """Convert to Polars DataFrame."""
        return (
            pl.read_json(BytesIO(orjson.dumps(self.dataframe_dict)))
            if self.dataframe_dict
            else self.dataframe
        )


def _combine_rules(
    rules: list[PseudoRule], target_rules: list[PseudoRule] | None
) -> list[tuple[PseudoRule, PseudoRule | None]]:
    combined: list[tuple[PseudoRule, PseudoRule | None]] = []

    # Zip rules and target_rules together; use None as target if target_rules is undefined
    for index, rule in enumerate(rules):
        combined.append((rule, target_rules[index] if target_rules else None))

    return combined


def _traverse_dataframe_dict(
    items: list[dict[str, t.Any] | None],
    rules: list[tuple[PseudoRule, PseudoRule | None]],
    metrics: Counter[str],
    prefix: str = "",
) -> t.Generator[FieldMatch, None, None]:
    for index, col in enumerate(items):
        if col is None:
            continue
        elif isinstance(col.get("datatype"), dict):
            next_prefix = (
                f"{prefix}[{index}]" if col["name"] == "" else f"{prefix}/{col['name']}"
            )
            yield from _traverse_dataframe_dict(
                col["values"], rules, metrics, next_prefix
            )
        elif len(col["values"]) > 0 and any(v is not None for v in col["values"]):
            name = f"{prefix}/{col['name']}".lstrip("/")
            for rule, target_rule in rules:
                if _glob_matches(_strip_array_indices(name), rule.pattern):
                    metrics.update({name: 1})
                    yield FieldMatch(
                        path=name,
                        col=col,
                        func=rule.func,
                        target_func=target_rule.func if target_rule else None,
                        pattern=rule.pattern,
                    )
                    break
    # print(_glob_matches.cache_info())


def _strip_array_indices(name: str) -> str:
    """Remove array index from the given path, e.g. remove the [1] part of /some/array[1]/struct/element.

    The array index is not used in glob matching anyway, and the result can be cached regardless of the array index
    """
    return ARRAY_INDEX_MATCHER.sub("", name)


@cache
def _glob_matches(name: str, rule: str) -> bool:
    """Use glob matching and cache the result."""
    return glob.globmatch(name.lower(), rule.lower(), flags=glob.GLOBSTAR | glob.BRACE)
